#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["OperationApiNameVisitor"]

from neptune.common.exceptions import InternalClientError
from neptune.internal.operation import (
    AddStrings,
    AssignArtifact,
    AssignBool,
    AssignDatetime,
    AssignFloat,
    AssignInt,
    AssignString,
    ClearArtifact,
    ClearFloatLog,
    ClearImageLog,
    ClearStringLog,
    ClearStringSet,
    ConfigFloatSeries,
    CopyAttribute,
    DeleteAttribute,
    DeleteFiles,
    LogFloats,
    LogImages,
    LogStrings,
    Operation,
    RemoveStrings,
    TrackFilesToArtifact,
    UploadFile,
    UploadFileContent,
    UploadFileSet,
)
from neptune.internal.operation_visitor import (
    OperationVisitor,
    Ret,
)


class OperationApiNameVisitor(OperationVisitor[str]):
    def visit(self, op: Operation) -> str:
        return op.accept(self)

    def visit_assign_float(self, _: AssignFloat) -> str:
        return "assignFloat"

    def visit_assign_int(self, _: AssignInt) -> str:
        return "assignInt"

    def visit_assign_bool(self, _: AssignBool) -> str:
        return "assignBool"

    def visit_assign_string(self, _: AssignString) -> str:
        return "assignString"

    def visit_assign_datetime(self, _: AssignDatetime) -> Ret:
        return "assignDatetime"

    def visit_upload_file(self, _: UploadFile) -> str:
        raise InternalClientError("Specialized endpoint should be used to upload file attribute")

    def visit_upload_file_content(self, _: UploadFileContent) -> str:
        raise InternalClientError("Specialized endpoint should be used to upload file attribute")

    def visit_upload_file_set(self, op: UploadFileSet) -> Ret:
        raise InternalClientError("Specialized endpoints should be used to upload file set attribute")

    def visit_log_floats(self, _: LogFloats) -> str:
        return "logFloats"

    def visit_log_strings(self, _: LogStrings) -> str:
        return "logStrings"

    def visit_log_images(self, _: LogImages) -> str:
        return "logImages"

    def visit_clear_float_log(self, _: ClearFloatLog) -> str:
        return "clearFloatSeries"

    def visit_clear_string_log(self, _: ClearStringLog) -> str:
        return "clearStringSeries"

    def visit_clear_image_log(self, _: ClearImageLog) -> str:
        return "clearImageSeries"

    def visit_config_float_series(self, _: ConfigFloatSeries) -> str:
        return "configFloatSeries"

    def visit_add_strings(self, _: AddStrings) -> str:
        return "insertStrings"

    def visit_remove_strings(self, _: RemoveStrings) -> str:
        return "removeStrings"

    def visit_delete_attribute(self, _: DeleteAttribute) -> str:
        return "deleteAttribute"

    def visit_clear_string_set(self, _: ClearStringSet) -> str:
        return "clearStringSet"

    def visit_delete_files(self, _: DeleteFiles) -> Ret:
        return "deleteFiles"

    def visit_assign_artifact(self, _: AssignArtifact) -> Ret:
        return "assignArtifact"

    def visit_track_files_to_artifact(self, _: TrackFilesToArtifact) -> Ret:
        raise InternalClientError("Specialized endpoint should be used to track artifact files")

    def visit_clear_artifact(self, _: ClearArtifact) -> Ret:
        return "clearArtifact"

    def visit_copy_attribute(self, _: CopyAttribute) -> Ret:
        raise NotImplementedError("This operation is client-side only")



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["get_boto_s3_client"]

import os

import boto3

from neptune.envs import S3_ENDPOINT_URL


def get_boto_s3_client():
    """
    User might want to use other than `AWS` `S3` providers, so we should be able to override `endpoint_url`.
    Unfortunately `boto3` doesn't support this parameter in configuration, so we'll have to create our env variable.
    boto3 supported config envs:
     * https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html#using-environment-variables
    boto3 `endpoint_url` support PR:
     * https://github.com/boto/boto3/pull/2746
    """
    endpoint_url = os.getenv(S3_ENDPOINT_URL)
    return boto3.resource(
        service_name="s3",
        endpoint_url=endpoint_url,
    )



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["OfflineNeptuneBackend"]

from typing import List

from neptune.exceptions import NeptuneOfflineModeFetchException
from neptune.internal.artifacts.types import ArtifactFileData
from neptune.internal.backends.api_model import (
    ArtifactAttribute,
    Attribute,
    BoolAttribute,
    DatetimeAttribute,
    FileAttribute,
    FloatAttribute,
    FloatSeriesAttribute,
    FloatSeriesValues,
    ImageSeriesValues,
    IntAttribute,
    StringAttribute,
    StringSeriesAttribute,
    StringSeriesValues,
    StringSetAttribute,
)
from neptune.internal.backends.neptune_backend_mock import NeptuneBackendMock
from neptune.internal.container_type import ContainerType


class OfflineNeptuneBackend(NeptuneBackendMock):
    WORKSPACE_NAME = "offline"

    def get_attributes(self, container_id: str, container_type: ContainerType) -> List[Attribute]:
        raise NeptuneOfflineModeFetchException

    def get_float_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> FloatAttribute:
        raise NeptuneOfflineModeFetchException

    def get_int_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> IntAttribute:
        raise NeptuneOfflineModeFetchException

    def get_bool_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> BoolAttribute:
        raise NeptuneOfflineModeFetchException

    def get_file_attribute(self, container_id: str, container_type: ContainerType, path: List[str]) -> FileAttribute:
        raise NeptuneOfflineModeFetchException

    def get_string_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> StringAttribute:
        raise NeptuneOfflineModeFetchException

    def get_datetime_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> DatetimeAttribute:
        raise NeptuneOfflineModeFetchException

    def get_artifact_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> ArtifactAttribute:
        raise NeptuneOfflineModeFetchException

    def list_artifact_files(self, project_id: str, artifact_hash: str) -> List[ArtifactFileData]:
        raise NeptuneOfflineModeFetchException

    def get_float_series_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> FloatSeriesAttribute:
        raise NeptuneOfflineModeFetchException

    def get_string_series_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> StringSeriesAttribute:
        raise NeptuneOfflineModeFetchException

    def get_string_set_attribute(
        self, container_id: str, container_type: ContainerType, path: List[str]
    ) -> StringSetAttribute:
        raise NeptuneOfflineModeFetchException

    def get_string_series_values(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        offset: int,
        limit: int,
    ) -> StringSeriesValues:
        raise NeptuneOfflineModeFetchException

    def get_float_series_values(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        offset: int,
        limit: int,
    ) -> FloatSeriesValues:
        raise NeptuneOfflineModeFetchException

    def get_image_series_values(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        offset: int,
        limit: int,
    ) -> ImageSeriesValues:
        raise NeptuneOfflineModeFetchException

    def download_file_series_by_index(
        self,
        container_id: str,
        container_type: ContainerType,
        path: List[str],
        index: int,
        destination: str,
    ):
        raise NeptuneOfflineModeFetchException



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = (
    "ProjectVisibility",
    "ProjectMemberRole",
    "MemberRole",
    "WorkspaceMemberRole",
)


class ProjectVisibility:
    PRIVATE = "priv"
    PUBLIC = "pub"
    WORKSPACE = "workspace"


class ProjectMemberRole:
    VIEWER = "viewer"
    OWNER = "owner"
    CONTRIBUTOR = "contributor"

    # Deprecated
    MEMBER = CONTRIBUTOR
    MANAGER = OWNER


MemberRole = ProjectMemberRole


class WorkspaceMemberRole:
    ADMIN = "admin"
    MEMBER = "member"



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["version", "__version__"]

import sys
from typing import Optional

from packaging.version import parse

from neptune.common.warnings import warn_once

if sys.version_info >= (3, 8):
    from importlib.metadata import PackageNotFoundError
    from importlib.metadata import version as version_parser
else:
    from importlib_metadata import PackageNotFoundError
    from importlib_metadata import version as version_parser


def check_version(package_name: str) -> Optional[str]:
    try:
        return version_parser(package_name)
    except PackageNotFoundError:
        # package is not installed
        return None


def detect_version() -> str:
    neptune_version = check_version("neptune")
    neptune_client_version = check_version("neptune-client")

    if neptune_version is not None and neptune_client_version is not None:
        raise RuntimeError(
            "We've detected that the 'neptune' and 'neptune-client' packages are both installed. "
            "Uninstall each of them and then install only the new 'neptune' package. For more, "
            "see https://docs.neptune.ai/setup/upgrading/"
        )
    elif neptune_version is not None:
        return neptune_version
    elif neptune_client_version is not None:
        warn_once(
            "The 'neptune-client' package has been deprecated and will be removed in the future. Install "
            "the 'neptune' package instead. For more, see https://docs.neptune.ai/setup/upgrading/"
        )
        return neptune_client_version

    raise PackageNotFoundError("neptune")


__version__ = detect_version()
version = parse(__version__)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["LocalFileHashStorage"]

import os
import sqlite3 as sql
from dataclasses import dataclass
from pathlib import Path


class LocalFileHashStorage:
    @dataclass
    class LocalFileHash:
        file_path: str
        file_hash: str
        modification_date: str

    def __init__(self):
        db_path = Path.home() / ".neptune" / "files.db"
        os.makedirs(db_path.parent, exist_ok=True)

        self.session = sql.connect(str(db_path))
        self.cursor: sql.Cursor = self.session.cursor()
        self.cursor.execute(
            "CREATE TABLE IF NOT EXISTS local_file_hashes (file_path text, file_hash text, modification_date text)"
        )
        self.session.commit()

    def insert(self, path: Path, computed_hash: str, modification_date: str):
        self.cursor.execute(
            "INSERT INTO local_file_hashes (file_path, file_hash, modification_date) VALUES (?, ?, ?)",
            (str(path), computed_hash, modification_date),
        )
        self.session.commit()

    def fetch_one(self, path: Path) -> "LocalFileHash":
        found = [
            LocalFileHashStorage.LocalFileHash(*row)
            for row in self.cursor.execute(
                "SELECT file_path, file_hash, modification_date FROM local_file_hashes WHERE file_path = ?",
                (str(path),),
            )
        ]

        return found[0] if found is not None and len(found) > 0 else None

    def update(self, path: Path, computed_hash: str, modification_date: str):
        self.cursor.execute(
            "UPDATE local_file_hashes SET file_hash=?, modification_date=? WHERE file_path = ?",
            (computed_hash, modification_date, str(path)),
        )
        self.session.commit()



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ("OfflineOperationProcessor",)

import threading
from pathlib import Path
from typing import Optional

from neptune.constants import (
    NEPTUNE_DATA_DIRECTORY,
    OFFLINE_DIRECTORY,
)
from neptune.internal.container_type import ContainerType
from neptune.internal.disk_queue import DiskQueue
from neptune.internal.id_formats import UniqueId
from neptune.internal.operation import Operation
from neptune.internal.operation_processors.operation_processor import OperationProcessor
from neptune.internal.operation_processors.operation_storage import OperationStorage


class OfflineOperationProcessor(OperationProcessor):
    def __init__(self, container_id: UniqueId, container_type: ContainerType, lock: threading.RLock):
        self._operation_storage = OperationStorage(self._init_data_path(container_id, container_type))

        self._queue = DiskQueue(
            dir_path=self._operation_storage.data_path,
            to_dict=lambda x: x.to_dict(),
            from_dict=Operation.from_dict,
            lock=lock,
        )

    @staticmethod
    def _init_data_path(container_id: UniqueId, container_type: ContainerType) -> Path:
        return Path(f"{NEPTUNE_DATA_DIRECTORY}/{OFFLINE_DIRECTORY}/{container_type.create_dir_name(container_id)}")

    def enqueue_operation(self, op: Operation, *, wait: bool) -> None:
        self._queue.put(op)

    def wait(self):
        self.flush()

    def flush(self):
        self._queue.flush()

    def start(self):
        pass

    def stop(self, seconds: Optional[float] = None):
        self._queue.close()



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["CompatibilityImporter"]

import sys
from importlib import import_module
from importlib.abc import (
    Loader,
    MetaPathFinder,
)


class CompatibilityModuleLoader(Loader):
    def module_repr(self, module):
        return repr(module)

    def load_module(self, fullname):
        module_name_parts = fullname.split(".")
        new_module_name = f"neptune.{module_name_parts[2]}"

        module = sys.modules[fullname] = import_module(new_module_name)
        return module


modules = [
    "neptune.new.attributes",
    "neptune.new.cli",
    "neptune.new.integrations",
    "neptune.new.logging",
    "neptune.new.metadata_containers",
    "neptune.new.types",
]


class CompatibilityImporter(MetaPathFinder):
    def find_module(self, fullname, path=None):
        if ".".join(fullname.split(".")) in modules:
            return CompatibilityModuleLoader()



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from abc import (
    ABC,
    abstractmethod,
)
from typing import Dict

from neptune.legacy.model import ChannelWithLastValue


class ApiClient(ABC):
    @property
    @abstractmethod
    def api_address(self):
        pass

    @property
    @abstractmethod
    def display_address(self):
        pass

    @property
    @abstractmethod
    def proxies(self):
        pass


class BackendApiClient(ApiClient, ABC):
    @abstractmethod
    def get_project(self, project_qualified_name):
        pass

    @abstractmethod
    def get_projects(self, namespace):
        pass

    @abstractmethod
    def create_leaderboard_backend(self, project) -> "LeaderboardApiClient":
        pass


class LeaderboardApiClient(ApiClient, ABC):
    @abstractmethod
    def get_project_members(self, project_identifier):
        pass

    @abstractmethod
    def get_leaderboard_entries(
        self,
        project,
        entry_types=None,
        ids=None,
        states=None,
        owners=None,
        tags=None,
        min_running_time=None,
    ):
        pass

    def websockets_factory(self, project_id, experiment_id):
        return None

    @abstractmethod
    def get_channel_points_csv(self, experiment, channel_internal_id, channel_name):
        pass

    @abstractmethod
    def get_metrics_csv(self, experiment):
        pass

    @abstractmethod
    def create_experiment(
        self,
        project,
        name,
        description,
        params,
        properties,
        tags,
        abortable,
        monitored,
        git_info,
        hostname,
        entrypoint,
        notebook_id,
        checkpoint_id,
    ):
        pass

    @abstractmethod
    def upload_source_code(self, experiment, source_target_pairs):
        pass

    @abstractmethod
    def get_notebook(self, project, notebook_id):
        pass

    @abstractmethod
    def get_last_checkpoint(self, project, notebook_id):
        pass

    @abstractmethod
    def create_notebook(self, project):
        pass

    @abstractmethod
    def create_checkpoint(self, notebook_id, jupyter_path, _file=None):
        pass

    @abstractmethod
    def get_experiment(self, experiment_id):
        pass

    @abstractmethod
    def set_property(self, experiment, key, value):
        pass

    @abstractmethod
    def remove_property(self, experiment, key):
        pass

    @abstractmethod
    def update_tags(self, experiment, tags_to_add, tags_to_delete):
        pass

    @abstractmethod
    def create_channel(self, experiment, name, channel_type) -> ChannelWithLastValue:
        pass

    @abstractmethod
    def get_channels(self, experiment) -> Dict[str, object]:
        pass

    @abstractmethod
    def reset_channel(self, experiment, channel_id, channel_name, channel_type):
        pass

    @abstractmethod
    def create_system_channel(self, experiment, name, channel_type) -> ChannelWithLastValue:
        pass

    @abstractmethod
    def get_system_channels(self, experiment) -> Dict[str, object]:
        pass

    @abstractmethod
    def send_channels_values(self, experiment, channels_with_values):
        pass

    @abstractmethod
    def mark_failed(self, experiment, traceback):
        pass

    @abstractmethod
    def ping_experiment(self, experiment):
        pass

    @abstractmethod
    def create_hardware_metric(self, experiment, metric):
        pass

    @abstractmethod
    def send_hardware_metric_reports(self, experiment, metrics, metric_reports):
        pass

    @abstractmethod
    def log_artifact(self, experiment, artifact, destination=None):
        pass

    @abstractmethod
    def delete_artifacts(self, experiment, path):
        pass

    @abstractmethod
    def download_data(self, experiment, path, destination):
        pass

    @abstractmethod
    def download_sources(self, experiment, path=None, destination_dir=None):
        pass

    @abstractmethod
    def download_artifacts(self, experiment, path=None, destination_dir=None):
        pass

    @abstractmethod
    def download_artifact(self, experiment, path=None, destination_dir=None):
        pass



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["Series"]

import abc
from typing import (
    Collection,
    Generic,
    Iterable,
    List,
    Optional,
    TypeVar,
    Union,
)

from neptune.attributes.attribute import Attribute
from neptune.internal.operation import LogOperation
from neptune.internal.types.stringify_value import StringifyValue
from neptune.internal.utils import (
    is_collection,
    is_stringify_value,
    verify_collection_type,
    verify_type,
)
from neptune.internal.utils.iteration import get_batches
from neptune.types.series.series import Series as SeriesVal

ValTV = TypeVar("ValTV", bound=SeriesVal)
DataTV = TypeVar("DataTV")
LogOperationTV = TypeVar("LogOperationTV", bound=LogOperation)


class Series(Attribute, Generic[ValTV, DataTV, LogOperationTV]):
    def __init_subclass__(cls, max_batch_size: int, operation_cls: type(LogOperationTV)):
        cls.max_batch_size = max_batch_size
        cls.operation_cls = operation_cls

    def clear(self, *, wait: bool = False) -> None:
        self._clear_impl(wait)

    def _get_log_operations_from_value(self, value: ValTV) -> List[LogOperationTV]:
        mapped_values = self._map_series_val(value)
        values_with_step_and_ts = zip(mapped_values, value.steps, value.timestamps)
        log_values = [self.operation_cls.ValueType(val, step=step, ts=ts) for val, step, ts in values_with_step_and_ts]
        return [
            self.operation_cls(self._path, chunk) for chunk in get_batches(log_values, batch_size=self.max_batch_size)
        ]

    @classmethod
    def _map_series_val(cls, value: ValTV) -> List[DataTV]:
        return value.values

    def _get_config_operation_from_value(self, value: ValTV) -> Optional[LogOperationTV]:
        return None

    @abc.abstractmethod
    def _get_clear_operation(self) -> LogOperationTV:
        pass

    @abc.abstractmethod
    def _data_to_value(self, values: Iterable, **kwargs) -> ValTV:
        pass

    @abc.abstractmethod
    def _is_value_type(self, value) -> bool:
        pass

    def _handle_stringified_value(self, value):
        return value.value

    def assign(self, value, wait: bool = False) -> None:
        if not self._is_value_type(value):
            value = self._data_to_value(value)
        clear_op = self._get_clear_operation()
        config_op = self._get_config_operation_from_value(value)
        with self._container.lock():
            if config_op:
                self._enqueue_operation(config_op, wait=False)
            if not value.values:
                self._enqueue_operation(clear_op, wait=wait)
            else:
                self._enqueue_operation(clear_op, wait=False)
                ops = self._get_log_operations_from_value(value)
                for op in ops:
                    self._enqueue_operation(op, wait=wait)

    def log(
        self,
        value: Union[DataTV, Iterable[DataTV]],
        step: Optional[float] = None,
        timestamp: Optional[float] = None,
        wait: bool = False,
        **kwargs,
    ) -> None:
        """log is a deprecated method, this code should be removed in future"""
        if step is not None:
            verify_type("step", step, (float, int))
        if timestamp is not None:
            verify_type("timestamp", timestamp, (float, int))

        if is_stringify_value(value):
            value = self._handle_stringified_value(value)

        if is_collection(value):
            if step is not None and len(value) > 1:
                raise ValueError("Collection of values is not supported for explicitly defined 'step'.")
            steps = None if step is None else [step] * len(value)
            timestamps = None if timestamp is None else [timestamp] * len(value)
            value = self._data_to_value(value, steps=steps, timestamps=timestamps, **kwargs)
        else:
            steps = None if step is None else [step]
            timestamps = None if timestamp is None else [timestamp]
            value = self._data_to_value([value], steps=steps, timestamps=timestamps, **kwargs)

        ops = self._get_log_operations_from_value(value)

        with self._container.lock():
            for op in ops:
                self._enqueue_operation(op, wait=wait)

    def extend(
        self,
        values: Union[Collection[DataTV], StringifyValue],
        steps: Optional[Collection[float]] = None,
        timestamps: Optional[Collection[float]] = None,
        wait: bool = False,
        **kwargs,
    ) -> None:
        if is_stringify_value(values):
            values = self._handle_stringified_value(values)

        if steps is not None:
            verify_collection_type("steps", steps, (float, int))
            if len(steps) != len(values):
                raise ValueError(f"Number of steps must be equal to number of values ({len(steps)} != {len(values)}")

        if timestamps is not None:
            verify_collection_type("timestamps", timestamps, (float, int))
            if len(timestamps) != len(values):
                raise ValueError(
                    f"Number of timestamps must be equal to number of values ({len(timestamps)} != {len(values)}"
                )

        value = self._data_to_value(values, steps=steps, timestamps=timestamps, **kwargs)
        ops = self._get_log_operations_from_value(value)

        with self._container.lock():
            for op in ops:
                self._enqueue_operation(op, wait=wait)

    def _clear_impl(self, wait: bool = False) -> None:
        op = self._get_clear_operation()
        with self._container.lock():
            self._enqueue_operation(op, wait=wait)



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["Credentials"]

import base64
import json
import os
from dataclasses import dataclass
from typing import (
    Dict,
    Optional,
)

from neptune.common.envs import API_TOKEN_ENV_NAME
from neptune.common.exceptions import NeptuneInvalidApiTokenException
from neptune.constants import ANONYMOUS_API_TOKEN
from neptune.exceptions import NeptuneMissingApiTokenException
from neptune.internal.constants import ANONYMOUS_API_TOKEN_CONTENT


@dataclass(frozen=True)
class Credentials:
    api_token: str
    token_origin_address: str
    api_url_opt: str

    @classmethod
    def from_token(cls, api_token: Optional[str] = None) -> "Credentials":
        if api_token is None:
            api_token = os.getenv(API_TOKEN_ENV_NAME)

        if api_token == ANONYMOUS_API_TOKEN:
            api_token = ANONYMOUS_API_TOKEN_CONTENT

        if api_token is None:
            raise NeptuneMissingApiTokenException()

        api_token = api_token.strip()
        token_dict = Credentials._api_token_to_dict(api_token)
        # TODO: Consider renaming 'api_address' (breaking backward compatibility)
        if "api_address" not in token_dict:
            raise NeptuneInvalidApiTokenException()
        token_origin_address = token_dict["api_address"]
        api_url = token_dict["api_url"] if "api_url" in token_dict else None

        return Credentials(
            api_token=api_token,
            token_origin_address=token_origin_address,
            api_url_opt=api_url,
        )

    @staticmethod
    def _api_token_to_dict(api_token: str) -> Dict[str, str]:
        try:
            return json.loads(base64.b64decode(api_token.encode()).decode("utf-8"))
        except Exception:
            raise NeptuneInvalidApiTokenException()



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["Set"]

import abc
from typing import (
    TYPE_CHECKING,
    TypeVar,
)

from neptune.types.value import Value

if TYPE_CHECKING:
    from neptune.types.value_visitor import ValueVisitor

Ret = TypeVar("Ret")


class Set(Value):
    @abc.abstractmethod
    def accept(self, visitor: "ValueVisitor[Ret]") -> Ret:
        pass



#
# Copyright (c) 2022, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["get_operation_processor"]

import threading

from neptune.internal.backends.neptune_backend import NeptuneBackend
from neptune.internal.container_type import ContainerType
from neptune.internal.id_formats import UniqueId
from neptune.types.mode import Mode

from .async_operation_processor import AsyncOperationProcessor
from .offline_operation_processor import OfflineOperationProcessor
from .operation_processor import OperationProcessor
from .read_only_operation_processor import ReadOnlyOperationProcessor
from .sync_operation_processor import SyncOperationProcessor


def get_operation_processor(
    mode: Mode,
    container_id: UniqueId,
    container_type: ContainerType,
    backend: NeptuneBackend,
    lock: threading.RLock,
    flush_period: float,
) -> OperationProcessor:
    if mode == Mode.ASYNC:
        return AsyncOperationProcessor(
            container_id,
            container_type,
            backend,
            lock,
            sleep_time=flush_period,
        )
    elif mode == Mode.SYNC:
        return SyncOperationProcessor(container_id, container_type, backend)
    elif mode == Mode.DEBUG:
        return SyncOperationProcessor(container_id, container_type, backend)
    elif mode == Mode.OFFLINE:
        # the object was returned by mocked backend and has some random ID.
        return OfflineOperationProcessor(container_id, container_type, lock)
    elif mode == Mode.READ_ONLY:
        return ReadOnlyOperationProcessor(container_id, backend)
    else:
        raise ValueError(f"mode should be one of {[m for m in Mode]}")
